"""
Zurich Instruments LabOne Python API Utility Functions.

This module provides basic utility functions for:

- Creating an API session by connecting to an appropriate Data Server.

- Detecting devices.

- Loading and saving device settings.

- Loading data saved by either the Zurich Instruments LabOne User Interface or
  ziControl into Python as numpy structured arrays.
"""

# Copyright 2014 Zurich Instruments AG.

import re
import warnings
import os
import time
import numpy as np
try:
    # load_labone_mat() requires scipy.io.loadmat()
    import scipy.io
except ImportError:
    # No fallback. No complaints upon importing zhinst.utils, handle/raise
    # exception when the function load_labone_mat() is called.
    pass

import zhinst.ziPython as ziPython


def autoDetect(daq, exclude=None):
    """
    Return a string containing the first device ID (not in the exclude list)
    that is attached to the Data Server connected via daq, an instance of the
    ziPython.ziDAQServer class.

    Args:

      daq (ziDAQServer): An instance of the ziPython.ziDAQServer class
        (representing an API session connected to a Data Server).

      exclude (list of str, optional): A list of strings specifying devices to
        exclude. autoDetect() will not return the name of a device in this
        list.

    Returns:

      A string specifying the first device ID not in exclude.

    Raises:

      RunTimeError: If no device was found.
      RunTimeError: If daq is not an instance of ziPython.ziDAQServer.

    Example:

      zhinst.utils
      daq = zhinst.utils.autoConnect()
      device = zhinst.utils.autoDetect(daq)
    """
    if not isinstance(daq, ziPython.ziDAQServer):
        raise RuntimeError("First argument must be an instance of ziPython.ziDAQServer")
    nodes = daq.listNodes('/', 0)
    devs = [node for node in nodes if re.match("dev*", node, re.IGNORECASE)]
    if exclude is None:
        exclude = []
    if not isinstance(exclude, list):
        exclude = [exclude]
    exclude = [x.lower() for x in exclude]
    devs = [dev for dev in devs if dev.lower() not in exclude]
    if len(devs) == 0:
        raise RuntimeError("No Device found. Make sure that the device is connected to the host via USB or Ethernet" +
                           "and that it is switched on. It may also be necessary to issue a connectDevice command.")
    # Found at least one device -> selection valid.
    # Select the first one
    device = devs[0].lower()
    print "autoDetect selected the device %s for the measurement." % device
    return device


def devices(daq):
    """
    Return a list of strings containing the device IDs that are attached to the
    Data Server connected via daq, an instance of the ziPython.ziDAQServer
    class. Returns an empty list if no devices are found.

    Args:

      daq (ziDAQServer): An instance of the ziPython.ziDAQServer class
        (representing an API session connected to a Data Server).

    Returns:

      A list of strings of connected device IDs. The list is empty if no devices
      are detected.

    Raises:

      RunTimeError: If daq is not an instance of ziPython.ziDAQServer.

    Example:

      import zhinst.utils
      daq = zhinst.utils.autoConnect()  # autoConnect not supported for MFLI devices
      device = zhinst.utils.autoDetect(daq)

    """
    if not isinstance(daq, ziPython.ziDAQServer):
        raise RuntimeError("First argument must be an instance of ziPython.ziDAQServer")
    nodes = daq.listNodes('/', 0)
    devs = [node for node in nodes if re.match("dev*", node, re.IGNORECASE)]
    devs = list(x.lower() for x in list(devs))
    return devs


def autoConnect(default_port=None, api_level=None):
    """
    Try to connect to a Zurich Instruments Data Server with an attached
    available UHF or HF2 device.

    Important: autoConnect() does not support MFLI devices.

    Args:

      default_port (int, optional): The default port to use when connecting to
        the Data Server (specify 8005 for the HF2 Data Server and 8004 for the
        UHF Data Server).

      api_level (int, optional): The API level to use, either 1, 4 or 5. HF2 only
        supports Level 1, Level 5 is recommended for UHF and MFLI devices.

    Returns:

      ziDAQServer: An instance of the ziPython.ziDAQServer class that is used
        for communication to the Data Server.

    Raises:

      RunTimeError: If no running Data Server is found or no device is found
        that is attached to a Data Server.x

    If default_port is not specified (=None) then first try to connect to a HF2,
    if no server devices are found then try to connect to an UHF. This behaviour
    is useful for the API examples. If we cannot connect to a server and/or
    detect a connected device raise a RunTimeError.

    If default_port is 8004 try to connect to a UHF; if it is 8005 try to
    connect to an HF2. If no server and device is detected on this port raise
    a RunTimeError.
    """
    if default_port is None:
        default_port = 8005
        secondary_port = 8004
    elif default_port in [8004, 8005]:
        # If a port is specified, then don't try to connect to a secondary port
        secondary_port = None
    else:
        error_msg = "autoConnect(): input argument default_port (%d) must be either 8004 or 8005." % default_port
        raise RuntimeError(error_msg)
    if api_level is None:
        # Note: level 1 used by default for both UHF and HF2, otherwise
        # backwards compatibility not maintained.
        api_level = 1

    port_device = {8005: 'HF2', 8004: 'UHFLI or MFLI'}
    port_valid_api_levels = {8005: [1], 8004: [1, 4, 5]}
    port_exception = {}
    try:
        assert api_level in port_valid_api_levels[default_port], \
            "Invalid API level (`%d`) specified for port %d (%s devices), valid API Levels: %s." \
            % (api_level, default_port, port_device[default_port], port_valid_api_levels[default_port])
        daq = ziPython.ziDAQServer('localhost', default_port, api_level)
        devs = devices(daq)
        assert len(devs) > 0, "Successfully connected to the server on port `%d`, API level `%d` but devices() \
returned an empty list: No devices are connected to this PC." % (default_port, api_level)
        # We have a server running and a device, we're done
        print "autoConnect connected to a server on port %d using API level %d." % (default_port, api_level)
        return daq
    except (RuntimeError, AssertionError) as e:
        port_exception[default_port] = e

    error_msg_no_dev = "Please ensure that the correct Zurich Instruments server is running for your device and that \
your device is connected to the server (try connecting first via the User Interface)."

    # If default_port is specified as an input argument, then secondary_port is
    # None. If we got here we had no success on default_port: raise an error.
    if secondary_port is None:
        error_msg = "autoConnect(): failed to connect to a running server or failed to find a device connected to the \
server on port %d (used for %s devices). %s The exception was: %s" \
            % (default_port, port_device[default_port], error_msg_no_dev, port_exception[default_port])
        raise RuntimeError(error_msg)

    try:
        assert api_level in port_valid_api_levels[secondary_port], \
            "Invalid API level specified for port %d (%s devices), valid API Levels: %s." \
            % (secondary_port, port_device[secondary_port], port_valid_api_levels[secondary_port])
        daq = ziPython.ziDAQServer('localhost', secondary_port, api_level)
        devs = devices(daq)
        assert len(devs) > 0, "Successfully connected to the server on port `%d`, API level `%d` but devices() \
returned an empty list: No devices are connected to this PC." % (secondary_port, api_level)
        # We have a server running and a device, we're done
        print "autoConnect connected to a server on port %d using API level %d." % (secondary_port, api_level)
        return daq
    except (RuntimeError, AssertionError):
        port_exception[secondary_port] = e

    # If we got here we failed to connect to a device. Raise a RunTimeError.
    error_msg = "autoConnect(): failed to connect to a running server or failed to find a device connected to the \
server. %s The exception on port %d (used for %s devices) was: %s The exception on port %d (used for %s devices) \
was: %s" % (error_msg_no_dev, default_port, port_device[default_port], port_exception[default_port],
            secondary_port, port_device[secondary_port], port_exception[secondary_port])
    raise RuntimeError(error_msg)


def get_default_settings_path(daq):
    """
    Return the default path used for settings by the ziDeviceSettings module.

    Arguments:

      daq (instance of ziDAQServer): A ziPython API session.

    Returns:

      settings_path (str): The default ziDeviceSettings path.
    """
    timeout = 500
    device_settings = daq.deviceSettings(timeout)
    settings_path = device_settings.get('devicesettings/path')['path'][0]
    device_settings.clear()
    return settings_path


def load_settings(daq, device, filename):
    """
    Load a LabOne settings file to the specified device. This function is
    synchronous; it will block until loading the settings has finished.

    Arguments:

      daq (instance of ziDAQServer): A ziPython API session.

      device (str): The device ID specifying where to load the settings,
      e.g., 'dev123'.

      filename (str): The filename of the xml settings file to load. The
      filename can include a relative or full path.

    Raises:

      RunTimeError: If loading the settings times out.

    Examples:

      import zhinst.utils as utils
      daq = utils.autoConnect()
      dev = utils.autoDetect(daq)

      # Then, e.g., load settings from a file in the current directory:
      utils.load_settings(daq, dev, 'my_settings.xml')
      # Then, e.g., load settings from the default LabOne settings path:
      filename = 'default_ui.xml'
      path = utils.get_default_settings_path(daq)
      utils.load_settings(daq, dev, path + os.sep + filename)
    """
    path, filename = os.path.split(filename)
    filename_noext = os.path.splitext(filename)[0]
    module_timeout = 500  # [ms]
    device_settings = daq.deviceSettings(module_timeout)
    device_settings.set('devicesettings/device', device)
    device_settings.set('devicesettings/filename', filename_noext)
    if path:
        device_settings.set('devicesettings/path', path)
    else:
        device_settings.set('devicesettings/path', '.' + os.sep)
    device_settings.set('devicesettings/command', 'load')
    try:
        device_settings.execute()
        t0 = time.time()
        timeout = 60
        while not device_settings.finished():
            time.sleep(0.05)
            if time.time() - t0 > timeout:
                raise RuntimeError("Unable to load device settings after %.f seconds." % timeout)
    finally:
        device_settings.clear()


def save_settings(daq, device, filename):
    """
    Save settings from the specified device to a LabOne settings file. This
    function is synchronous; it will block until saving the settings has
    finished.

    Arguments:

      daq (instance of ziDAQServer): A ziPython API session.

      device (str): The device ID specifying where to load the settings,
      e.g., 'dev123'.

      filename (str): The filename of the LabOne xml settings file. The filename
      can include a relative or full path.

    Raises:

      RunTimeError: If saving the settings times out.

    Examples:

      import zhinst.utils as utils
      daq = utils.autoConnect()
      dev = utils.autoDetect(daq)

      # Then, e.g., save settings to a file in the current directory:
      utils.save_settings(daq, dev, 'my_settings.xml')

      # Then, e.g., save settings to the default LabOne settings path:
      filename = 'my_settings_example.xml'
      path = utils.get_default_settings_path(daq)
      utils.save_settings(daq, dev, path + os.sep + filename)
    """
    path, filename = os.path.split(filename)
    filename_noext = os.path.splitext(filename)[0]
    module_timeout = 500  # [ms]
    device_settings = daq.deviceSettings(module_timeout)
    device_settings.set('devicesettings/device', device)
    device_settings.set('devicesettings/filename', filename_noext)
    if path:
        device_settings.set('devicesettings/path', path)
    else:
        device_settings.set('devicesettings/path', '.' + os.sep)
    device_settings.set('devicesettings/command', 'save')
    try:
        device_settings.execute()
        t0 = time.time()
        timeout = 60
        while not device_settings.finished():
            time.sleep(0.05)
            if time.time() - t0 > timeout:
                raise RuntimeError("Unable to save device settings after %.f seconds." % timeout)
    finally:
        device_settings.clear()


# The names correspond to the data in the columns of a CSV file saved by the
# LabOne User Interface. These are the names of demodulator sample fields.
LABONE_DEMOD_NAMES = ('chunk', 'timestamp', 'x', 'y', 'freq', 'phase', 'dio', 'trigger', 'auxin0', 'auxin1')
LABONE_DEMOD_FORMATS = ('u8', 'u8', 'f8', 'f8', 'f8', 'f8', 'u4', 'u4', 'f8', 'f8')
# The dtype to provide when creating a numpy array from LabOne demodulator data
LABONE_DEMOD_DTYPE = zip(LABONE_DEMOD_NAMES, LABONE_DEMOD_FORMATS)

# The names correspond to the data in the columns of a CSV file saved by the
# ziControl User Interface. These are the names of demodulator sample fields.
ZICONTROL_NAMES = ('t', 'x', 'y', 'freq', 'dio', 'auxin0', 'auxin1')
ZICONTROL_FORMATS = ('f8', 'f8', 'f8', 'f8', 'u4', 'f8', 'f8')
# The dtype to provide when creating a numpy array from ziControl-saved demodulator data
ZICONTROL_DTYPE = zip(ZICONTROL_NAMES, ZICONTROL_FORMATS)


def load_labone_demod_csv(fname, column_names=LABONE_DEMOD_NAMES):
    """
    Load a CSV file containing demodulator samples as saved by the LabOne User
    Interface into a numpy structured array.

    Arguments:

      fname (file or str): The file or filename of the CSV file to load.

      column_names (list or tuple of str, optional): A list (or tuple) of column
      names to load from the CSV file. Default is to load all columns.

    Returns:

      sample (numpy ndarray): A numpy structured array of shape (num_points,)
      whose field names correspond to the column names in the first line of the
      CSV file. num_points is the number of lines in the CSV file - 1.

    Example:

      import zhinst.utils
      sample = zhinst.utils.load_labone_demod_csv('dev2004_demods_0_sample_00000.csv', ('timestamp', 'x', 'y'))
      import matplotlib.pyplot as plt
      import numpy as np
      plt.plot(sample['timestamp'], np.abs(sample['x'] + 1j*sample['y']))
    """
    assert set(column_names).issubset(LABONE_DEMOD_NAMES), \
        'Invalid name in ``column_names``, valid names are: %s' % str(LABONE_DEMOD_NAMES)
    cols = [col for col, dtype in enumerate(LABONE_DEMOD_DTYPE) if dtype[0] in column_names]
    dtype = [dt for dt in LABONE_DEMOD_DTYPE if dt[0] in column_names]
    sample = np.genfromtxt(fname, delimiter=';', dtype=dtype, usecols=cols, skip_header=1)
    return sample


def load_labone_csv(fname):
    """
    Load a CSV file containing generic data as saved by the LabOne User
    Interface into a numpy structured array.

    Arguments:

      filename (str): The filename of the CSV file to load.

    Returns:

      sample (numpy ndarray): A numpy structured array of shape (num_points,)
      whose field names correspond to the column names in the first line of the
      CSV file. num_points is the number of lines in the CSV file - 1.

    Example:

      import zhinst.utils
      # Load the CSV file of PID error data (node: /dev2004/pids/0/error)
      data = zhinst.utils.load_labone_csv('dev2004_pids_0_error_00000.csv')
      import matplotlib.pyplot as plt
      # Plot the error
      plt.plot(data['timestamp'], data['value'])
    """
    data = np.genfromtxt(fname, delimiter=';', dtype=None, names=True)
    return data


def load_labone_mat(filename):
    """
    A wrapper function for loading a MAT file as saved by the LabOne User
    Interface with scipy.io's loadmat() function. This function is included
    mainly to document how to work with the data structure return by
    scipy.io.loadmat().

    Arguments:

      filename (str): the name of the MAT file to load.

    Returns:

      data (dict): a nested dictionary containing the instrument data as
      specified in the LabOne User Interface. The nested structure of ``data``
      corresponds to the path of the data's node in the instrument's node
      hierarchy.

    Further comments:

      The MAT file saved by the LabOne User Interface (UI) is a Matlab V5.0 data
      file. The LabOne UI saves the specified data using native Matlab data
      structures in the same format as are returned by commands in the LabOne
      Matlab API. More specifically, these data structures are nested Matlab
      structs, the nested structure of which correspond to the location of the
      data in the instrument's node hierarchy.

      Matlab structs are returned by scipy.io.loadmat() as dictionaries, the
      name of the struct becomes a key in the dictionary. However, as for all
      objects in MATLAB, structs are in fact arrays of structs, where a single
      struct is an array of shape (1, 1). This means that each (nested)
      dictionary that is returned (corresponding to a node in node hierarchy) is
      loaded by scipy.io.loadmat as a 1-by-1 array and must be indexed as
      such. See the ``Example`` section below.

      For more information please refer to the following link:
      http://docs.scipy.org/doc/scipy/reference/tutorial/io.html#matlab-structs

    Example:

      device = 'dev88'
      # See ``Further explanation`` above for a comment on the indexing:
      timestamp = data[device][0,0]['demods'][0,0]['sample'][0,0]['timestamp'][0]
      x = data[device][0,0]['demods'][0,0]['sample'][0,0]['x'][0]
      y = data[device][0,0]['demods'][0,0]['sample'][0,0]['y'][0]
      import matplotlib.pyplot as plt
      import numpy as np
      plt.plot(timestamp, np.abs(x + 1j*y))

      # If multiple demodulator's are saved, data from the second demodulator,
      # e.g., is accessed as following:
      x = data[device][0,0]['demods'][0,1]['sample'][0,0]['x'][0]
    """
    try:
        data = scipy.io.loadmat(filename)
        return data
    except (NameError, AttributeError):
        print "\n *** Please install the ``scipy`` package and verify you can use scipy.io.loadmat()" \
            + " in order to use zhinst.utils.load_labone_mat. *** \n"
        print "The following exception was thrown:"
        raise
    except:
        raise


def load_zicontrol_csv(filename, column_names=ZICONTROL_NAMES):
    """
    Load a CSV file containing demodulator samples as saved by the ziControl
    User Interface into a numpy structured array.

    Arguments:

      filename (str): The file or filename of the CSV file to load.

      column_names (list or tuple of str, optional): A list (or tuple) of column
      names (demodulator sample field names) to load from the CSV file. Default
      is to load all columns.

    Returns:

      sample (numpy ndarray): A numpy structured array of shape (num_points,)
      whose field names correspond to the field names of a ziControl demodulator
      sample. num_points is the number of lines in the CSV file - 1.

    Example:

      import zhinst.utils
      sample = zhinst.utils.load_labone_csv('Freq1.csv', ('t', 'x', 'y'))
      import matplotlib.plt as plt
      import numpy as np
      plt.plot(sample['t'], np.abs(sample['x'] + 1j*sample['y']))
    """
    assert set(column_names).issubset(ZICONTROL_NAMES), \
        'Invalid name in ``column_names``, valid names are: %s' % str(ZICONTROL_NAMES)
    cols = [col for col, dtype in enumerate(ZICONTROL_DTYPE) if dtype[0] in column_names]
    dtype = [dt for dt in ZICONTROL_DTYPE if dt[0] in column_names]
    sample = np.genfromtxt(filename, delimiter=',', dtype=dtype, usecols=cols)
    return sample


def load_zicontrol_zibin(filename, column_names=ZICONTROL_NAMES):
    """
    Load a ziBin file containing demodulator samples as saved by the ziControl
    User Interface into a numpy structured array. This is for data saved by
    ziControl in binary format.

    Arguments:

      filename (str): The filename of the .ziBin file to load.

      column_names (list or tuple of str, optional): A list (or tuple) of column
      names to load from the CSV file. Default is to load all columns.

    Returns:

      sample (numpy ndarray): A numpy structured array of shape (num_points,)
      whose field names correspond to the field names of a ziControl demodulator
      sample. num_points is the number of sample points saved in the file.

    Further comments:

      Specifying a fewer names in ``column_names`` will not result in a speed-up
      as all data is loaded from the binary file by default.

    Example:

      import zhinst.utils
      sample = zhinst.utils.load_zicontrol_zibin('Freq1.ziBin')
      import matplotlib.plt as plt
      import numpy as np
      plt.plot(sample['t'], np.abs(sample['x'] + 1j*sample['y']))
    """
    assert set(column_names).issubset(ZICONTROL_NAMES), \
        'Invalid name in ``column_names``, valid names are: %s.' % str(ZICONTROL_NAMES)
    sample = np.fromfile(filename, dtype='>f8')
    rem = np.size(sample) % len(ZICONTROL_NAMES)
    assert rem == 0, "Incorrect number of data points in ziBin file, " + \
        "the number of data points must be divisible by the number of demodulator fields."
    n = np.size(sample) / len(ZICONTROL_NAMES)
    sample = np.reshape(sample, (n, len(ZICONTROL_NAMES))).transpose()
    cols = [col for col, dtype in enumerate(ZICONTROL_DTYPE) if dtype[0] in column_names]
    dtype = [dt for dt in ZICONTROL_DTYPE if dt[0] in column_names]
    sample = np.core.records.fromarrays(sample[cols, :], dtype=dtype)
    return sample


def check_for_sampleloss(timestamps):
    """
    Check whether timestamps are equidistantly spaced, it not, it is an
    indication that sampleloss has occurred whilst recording the demodulator
    data.

    This function assumes that the timestamps originate from continuously saved
    demodulator data, during which the demodulator sampling rate was not
    changed.

    Arguments:

      timestamp (numpy array): a 1-dimensional array containing
      demodulator timestamps

    Returns:

      idx (numpy array): a 1-dimensional array indicating the indices in
      timestamp where sampleloss has occurred. An empty array is returned in no
      sampleloss was present.
    """
    # If the second difference of the timestamps is zero, no sampleloss has occurred
    idx = np.where(np.diff(timestamps, n=2) > 0.1)[0] + 1
    # Find the true dtimestamps (determined by the configured sampling rate)
    dtimestamp = np.nan
    for i in range(0, np.shape(timestamps)[0]):
        # Take the sampling rate from a point where sample loss has not
        # occurred.
        if i not in idx:
            dtimestamp = timestamps[i + 1] - timestamps[i]
            break
    assert not np.isnan(dtimestamp)
    for i in idx:
        warnings.warn("Sample loss detected at timestamps=%d (%d points)",
                      timestamps[i], (timestamps[i + 1] - timestamps[i])/dtimestamp)
    return idx
